#!/usr/bin/env python3
#
# Et forsøk på å bruke SVD til å komprimere bilder.
#
# Forfatter: John Aslak Wee Kleven
# Lisens: GPL 3.0 (https://www.gnu.org/licenses/gpl-3.0.en.html)

# Se her for mer:
# https://zerobone.net/blog/cs/svd-image-compression/
# https://timbaumann.info/svd-image-compression-demo/

import numpy as np
import matplotlib.pyplot as plt # https://matplotlib.org/2.0.2/users/image_tutorial.html
from matplotlib import cm
from PIL import Image
from math import ceil

# Innstillinger
filname = "fugl"
ext = ".jpg"
fname = filname + ext
stop_percents = [0.05, 0.5, 1, 2.5, 5, 10, 20, 30, 40]

def perform_svd_compression(U, S, Vt, stop_percent):

	m, n = np.shape(U)[0], np.shape(Vt)[0]

	# Prosentandelen av singulære verdier vi vil bevare.
	stopper = ceil(len(S) * stop_percent/100)


	S = S[:stopper]
	# Legg til nuller til S for å fikse problemer med numpys
	# np.diag, som kun lager kvadratiske matriser mens SVD trenger
	# at den ikke (nødvendigvis) er kvadratisk.
	S = np.pad(S, (0, ( max(m, n) - len(S) )))

	# denne er raskere
	new_matr = U[:,:stopper] @ np.diag(S)[:stopper,:stopper] @ Vt[:stopper,:]

	print("Recovering the matrix done")
	print(f"K is a {m}*{n} matrix.")
	print(f"Stop percent is {stop_percent}.")
	print(f"There are {len(S)} singular values. Stopping at {stopper}.")
	print(f"Datapoints used (k*(m+n+1)) = {stopper*(m+n+1)}, and m*n = {m*n}")
	print(f"Compression factor: {round((m*n) / (stopper*(m+n+1)), 4)}")

	return new_matr


def svd_create(stop_percents):

	A = Image.open(fname).convert("L") # gjør til sort-hvitt. Verdier mellom 0 og 255.
	A = np.asarray(A)

	print("Performing SVD")
	U, S, Vt = np.linalg.svd(A)
	# Merk at 'Vt' er den transponerte til 'V' slik det vanligvis er i SVD.
	# hent tilbake K som "U @ np.diag(S) @ Vt" hvis du bruker numpy.
	# Man kan hente tilbake A ved å beregne U @ X @ Vt
	# hvor 'X' er en matrise av riktig størrelse, med elementene
	# til S på diagonalen. Kan sikkert også bli gjort på en annen måte.
	print("SVD done")

	for stop_percent in stop_percents:

		B = perform_svd_compression(U, S, Vt, stop_percent)

		# Now, prepare B for the picture.
		B = np.asarray(B)
		B = np.maximum(np.minimum(B, 255), 0) # sørg for at verdiene er mellom 0 og 255
		B = B.astype(np.uint8)

		plt.imsave(filname + "_greyscale_svd_" + str(stop_percent) + ext, B, cmap=cm.gray)
		print("Saved " + filname + "_greyscale_svd_" + str(stop_percent) + ext)

	# Lag et plott av de singulære verdiene
	plt.bar(list(range(1, len(S)+1)), list(S))
	plt.title("Singulære verdier til " + fname)
	plt.savefig(filname + "_singular_values" + ext, dpi=1000)
	print("Saved " + filname + "_singular_values" + ext)
	plt.semilogy()
	plt.savefig(filname + "_singular_values_log" + ext, dpi=1000)
	print("Saved " + filname + "_singular_values_log" + ext)

	plt.imsave(filname + "_greyscale_orig" + ext, A, cmap=cm.gray)
	print("Saved " + filname + "_greyscale_orig" + ext)


svd_create(stop_percents)
